@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2009 Carnegie Mellon University
+ *  Copyright (c) 2007-2012 Carnegie Mellon University
  *  All rights reserved.
  *
  *  OpenSlide is free software: you can redistribute it and/or modify
@@ -28,7 +28,7 @@
 struct _openslide_cache_key {
   int64_t x;
   int64_t y;
-  int32_t layer;
+  int32_t level;
 };
 
 // hash table value
@@ -48,6 +48,7 @@ struct _openslide_cache_entry {
 };
 
 struct _openslide_cache {
+  GMutex *mutex;
   GQueue *list;
   GHashTable *hashtable;
 
@@ -56,6 +57,7 @@ struct _openslide_cache {
 };
 
 // eviction
+// mutex must be held
 static void possibly_evict(struct _openslide_cache *cache, int incoming_size) {
   g_assert(incoming_size >= 0);
 
@@ -64,8 +66,7 @@ static void possibly_evict(struct _openslide_cache *cache, int incoming_size) {
 
   while(size > target) {
     // get key of last element
-    struct _openslide_cache_value *value =
-      (struct _openslide_cache_value *) g_queue_peek_tail(cache->list);
+    struct _openslide_cache_value *value = g_queue_peek_tail(cache->list);
     if (value == NULL) {
       return; // cache is empty
     }
@@ -84,22 +85,22 @@ static void possibly_evict(struct _openslide_cache *cache, int incoming_size) {
 
 // hash function helpers
 static guint hash_func(gconstpointer key) {
-  const struct _openslide_cache_key *c_key = (const struct _openslide_cache_key *) key;
+  const struct _openslide_cache_key *c_key = key;
 
   // assume 32-bit hash
 
-  // take the top 4 bits for layer, then 14 bits per x and y,
+  // take the top 4 bits for level, then 14 bits per x and y,
   // xor it all together
-  return (guint) ((c_key->layer << 28) ^ (c_key->y << 14) ^ (c_key->x));
+  return (guint) ((c_key->level << 28) ^ (c_key->y << 14) ^ (c_key->x));
 }
 
 static gboolean key_equal_func(gconstpointer a,
 			       gconstpointer b) {
-  const struct _openslide_cache_key *c_a = (const struct _openslide_cache_key *) a;
-  const struct _openslide_cache_key *c_b = (const struct _openslide_cache_key *) b;
+  const struct _openslide_cache_key *c_a = a;
+  const struct _openslide_cache_key *c_b = b;
 
   return (c_a->x == c_b->x) && (c_a->y == c_b->y) &&
-    (c_a->layer == c_b->layer);
+    (c_a->level == c_b->level);
 }
 
 static void hash_destroy_key(gpointer data) {
@@ -107,7 +108,7 @@ static void hash_destroy_key(gpointer data) {
 }
 
 static void hash_destroy_value(gpointer data) {
-  struct _openslide_cache_value *value = (struct _openslide_cache_value *) data;
+  struct _openslide_cache_value *value = data;
 
   // remove the item from the list
   g_queue_delete_link(value->cache->list, value->link);
@@ -126,6 +127,9 @@ static void hash_destroy_value(gpointer data) {
 struct _openslide_cache *_openslide_cache_create(int capacity_in_bytes) {
   struct _openslide_cache *cache = g_slice_new0(struct _openslide_cache);
 
+  // init mutex
+  cache->mutex = g_mutex_new();
+
   // init queue
   cache->list = g_queue_new();
 
@@ -143,26 +147,36 @@ struct _openslide_cache *_openslide_cache_create(int capacity_in_bytes) {
 
 void _openslide_cache_destroy(struct _openslide_cache *cache) {
   // clear hashtable (auto-deletes all data)
+  g_mutex_lock(cache->mutex);
   g_hash_table_unref(cache->hashtable);
+  g_mutex_unlock(cache->mutex);
 
   // clear list
   g_queue_free(cache->list);
 
+  // free mutex
+  g_mutex_free(cache->mutex);
+
   // destroy struct
   g_slice_free(struct _openslide_cache, cache);
 }
 
 
 int _openslide_cache_get_capacity(struct _openslide_cache *cache) {
-  return cache->capacity;
+  g_mutex_lock(cache->mutex);
+  int capacity = cache->capacity;
+  g_mutex_unlock(cache->mutex);
+  return capacity;
 }
 
 void _openslide_cache_set_capacity(struct _openslide_cache *cache,
 				   int capacity_in_bytes) {
   g_assert(capacity_in_bytes >= 0);
 
+  g_mutex_lock(cache->mutex);
   cache->capacity = capacity_in_bytes;
   possibly_evict(cache, 0);
+  g_mutex_unlock(cache->mutex);
 }
 
 // put and get
@@ -172,7 +186,7 @@ void _openslide_cache_set_capacity(struct _openslide_cache *cache,
 void _openslide_cache_put(struct _openslide_cache *cache,
 			  int64_t x,
 			  int64_t y,
-			  int32_t layer,
+			  int32_t level,
 			  void *data,
 			  int size_in_bytes,
 			  struct _openslide_cache_entry **_entry) {
@@ -185,9 +199,13 @@ void _openslide_cache_put(struct _openslide_cache *cache,
   entry->size = size_in_bytes;
   *_entry = entry;
 
+  // lock
+  g_mutex_lock(cache->mutex);
+
   // don't try to put anything in the cache that cannot possibly fit
   if (size_in_bytes > cache->capacity) {
-    //g_debug("refused %p", (void *) entry);
+    //g_debug("refused %p", entry);
+    g_mutex_unlock(cache->mutex);
     return;
   }
 
@@ -197,7 +215,7 @@ void _openslide_cache_put(struct _openslide_cache *cache,
   struct _openslide_cache_key *key = g_slice_new(struct _openslide_cache_key);
   key->x = x;
   key->y = y;
-  key->layer = layer;
+  key->level = level;
 
   // create value
   struct _openslide_cache_value *value =
@@ -219,24 +237,30 @@ void _openslide_cache_put(struct _openslide_cache *cache,
   // another ref for the cache
   g_atomic_int_inc(&entry->refcount);
 
-  //g_debug("insert %p", (void *) entry);
+  // unlock
+  g_mutex_unlock(cache->mutex);
+
+  //g_debug("insert %p", entry);
 }
 
 // entry must be unreffed when the caller is done with the data
 void *_openslide_cache_get(struct _openslide_cache *cache,
 			   int64_t x,
 			   int64_t y,
-			   int32_t layer,
-			   struct _openslide_cache_entry **entry) {
+			   int32_t level,
+			   struct _openslide_cache_entry **_entry) {
+  // lock
+  g_mutex_lock(cache->mutex);
+
   // create key
-  struct _openslide_cache_key key = { x, y, layer };
+  struct _openslide_cache_key key = { .x = x, .y = y, .level = level };
 
   // lookup key, maybe return NULL
-  struct _openslide_cache_value *value =
-    (struct _openslide_cache_value *) g_hash_table_lookup(cache->hashtable,
-							  &key);
+  struct _openslide_cache_value *value = g_hash_table_lookup(cache->hashtable,
+							     &key);
   if (value == NULL) {
-    *entry = NULL;
+    g_mutex_unlock(cache->mutex);
+    *_entry = NULL;
     return NULL;
   }
 
@@ -246,19 +270,22 @@ void *_openslide_cache_get(struct _openslide_cache *cache,
   g_queue_push_head_link(cache->list, link);
 
   // acquire entry reference for the caller
-  g_atomic_int_inc(&value->entry->refcount);
+  struct _openslide_cache_entry *entry = value->entry;
+  g_atomic_int_inc(&entry->refcount);
 
-  //g_debug("cache hit! %p %"G_GINT64_FORMAT" %"G_GINT64_FORMAT" %d", (void *) value->entry, x, y, layer);
+  //g_debug("cache hit! %p %"G_GINT64_FORMAT" %"G_GINT64_FORMAT" %d", entry, x, y, level);
+
+  // unlock
+  g_mutex_unlock(cache->mutex);
 
   // return data
-  *entry = value->entry;
-  return value->entry->data;
+  *_entry = entry;
+  return entry->data;
 }
 
 // value unref
-// calls do not need to be serialized
 void _openslide_cache_entry_unref(struct _openslide_cache_entry *entry) {
-  //g_debug("unref %p, refs %d", (void *) entry, g_atomic_int_get(&entry->refcount));
+  //g_debug("unref %p, refs %d", entry, g_atomic_int_get(&entry->refcount));
 
   if (g_atomic_int_dec_and_test(&entry->refcount)) {
     // free the data
@@ -267,6 +294,6 @@ void _openslide_cache_entry_unref(struct _openslide_cache_entry *entry) {
     // free the entry
     g_slice_free(struct _openslide_cache_entry, entry);
 
-    //g_debug("free %p", (void *) entry);
+    //g_debug("free %p", entry);
   }
 }
