@@ -28,19 +28,15 @@
 
 #include <string.h>
 #include <stdint.h>
+#include <math.h>
 #include <glib.h>
-#ifndef _MSC_VER
-#include <inttypes.h>
-#endif
 #include <cairo.h>
 
+#define NGR_TILE_HEIGHT 64
 
 struct ngr_data {
   int32_t ngr_count;
   struct _openslide_ngr **ngrs;
-
-  // cache lock
-  GMutex *cache_mutex;
 };
 
 
@@ -53,62 +49,71 @@ static void destroy_ngrs(struct _openslide_ngr **ngrs, int count) {
 }
 
 static void destroy(openslide_t *osr) {
-  struct ngr_data *data = (struct ngr_data *) osr->data;
+  struct ngr_data *data = osr->data;
 
   destroy_ngrs(data->ngrs, data->ngr_count);
 
-  g_mutex_free(data->cache_mutex);
   g_slice_free(struct ngr_data, data);
 }
 
 
-static void get_dimensions(openslide_t *osr, int32_t layer,
+static void get_dimensions(openslide_t *osr, int32_t level,
 			   int64_t *w, int64_t *h) {
 
-  struct ngr_data *data = (struct ngr_data *) osr->data;
-  struct _openslide_ngr *ngr = data->ngrs[layer];
+  struct ngr_data *data = osr->data;
+  struct _openslide_ngr *ngr = data->ngrs[level];
 
   *w = ngr->w;
   *h = ngr->h;
 }
 
+static void get_tile_geometry(openslide_t *osr, int32_t level,
+                              int64_t *w, int64_t *h) {
+  struct ngr_data *data = osr->data;
+  struct _openslide_ngr *ngr = data->ngrs[level];
+
+  *w = ngr->column_width;
+  *h = NGR_TILE_HEIGHT;
+}
+
 static void read_tile(openslide_t *osr,
 		      cairo_t *cr,
-		      int32_t layer,
+		      int32_t level,
 		      int64_t tile_x, int64_t tile_y,
 		      double translate_x, double translate_y,
-		      struct _openslide_cache *cache) {
-  struct ngr_data *data = (struct ngr_data *) osr->data;
-  struct _openslide_ngr *ngr = data->ngrs[layer];
+		      struct _openslide_cache *cache,
+		      void *arg G_GNUC_UNUSED) {
+  struct ngr_data *data = osr->data;
+  struct _openslide_ngr *ngr = data->ngrs[level];
+  GError *tmp_err = NULL;
 
   // check if beyond boundary
   int num_columns = ngr->w / ngr->column_width;
-  if ((tile_x >= num_columns) || (tile_y >= ngr->h)) {
+  int num_rows = (ngr->h + NGR_TILE_HEIGHT - 1) / NGR_TILE_HEIGHT;
+  if (tile_x >= num_columns || tile_y >= num_rows) {
     return;
   }
 
-  int tilesize = ngr->column_width * 4;
+  int64_t tw = ngr->column_width;
+  int64_t th = MIN(NGR_TILE_HEIGHT, ngr->h - tile_y * NGR_TILE_HEIGHT);
+  int tilesize = tw * th * 4;
   struct _openslide_cache_entry *cache_entry;
   // look up tile in cache
-  g_mutex_lock(data->cache_mutex);
-  uint32_t *tiledata = (uint32_t *) _openslide_cache_get(cache,
-							 tile_x,
-							 tile_y,
-							 layer,
-							 &cache_entry);
-  g_mutex_unlock(data->cache_mutex);
+  uint32_t *tiledata = _openslide_cache_get(cache, tile_x, tile_y, level,
+                                            &cache_entry);
 
   if (!tiledata) {
     // read the tile data
-    FILE *f = _openslide_fopen(ngr->filename, "rb");
+    FILE *f = _openslide_fopen(ngr->filename, "rb", &tmp_err);
     if (!f) {
-      _openslide_set_error(osr, "Cannot open file %s", ngr->filename);
+      _openslide_set_error_from_gerror(osr, tmp_err);
+      g_clear_error(&tmp_err);
       return;
     }
 
     // compute offset to read
     int64_t offset = ngr->start_in_file +
-      (tile_y * ngr->column_width * 6) +
+      (tile_y * NGR_TILE_HEIGHT * ngr->column_width * 6) +
       (tile_x * ngr->h * ngr->column_width * 6);
     //    g_debug("tile_x: %" G_GINT64_FORMAT ", "
     //	    "tile_y: %" G_GINT64_FORMAT ", "
@@ -116,8 +121,8 @@ static void read_tile(openslide_t *osr,
     fseeko(f, offset, SEEK_SET);
 
     // alloc and read
-    int buf_size = ngr->column_width * 6;
-    uint16_t *buf = (uint16_t *) g_slice_alloc(buf_size);
+    int buf_size = tw * th * 6;
+    uint16_t *buf = g_slice_alloc(buf_size);
 
     if (fread(buf, buf_size, 1, f) != 1) {
       _openslide_set_error(osr, "Cannot read file %s", ngr->filename);
@@ -128,8 +133,8 @@ static void read_tile(openslide_t *osr,
     fclose(f);
 
     // got the data, now convert to 8-bit xRGB
-    tiledata = (uint32_t *) g_slice_alloc(tilesize);
-    for (int i = 0; i < ngr->column_width; i++) {
+    tiledata = g_slice_alloc(tilesize);
+    for (int i = 0; i < tw * th; i++) {
       // scale down from 12 bits
       uint8_t r = GINT16_FROM_LE(buf[(i * 3)]) >> 4;
       uint8_t g = GINT16_FROM_LE(buf[(i * 3) + 1]) >> 4;
@@ -140,19 +145,16 @@ static void read_tile(openslide_t *osr,
     g_slice_free1(buf_size, buf);
 
     // put it in the cache
-    g_mutex_lock(data->cache_mutex);
-    _openslide_cache_put(cache, tile_x, tile_y, layer,
+    _openslide_cache_put(cache, tile_x, tile_y, level,
                          tiledata,
-                         ngr->column_width * 1 * 4,
+                         tilesize,
                          &cache_entry);
-    g_mutex_unlock(data->cache_mutex);
   }
 
   // draw it
-  int64_t tw = ngr->column_width;
   cairo_surface_t *surface = cairo_image_surface_create_for_data((unsigned char *) tiledata,
 								 CAIRO_FORMAT_RGB24,
-								 tw, 1,
+								 tw, th,
 								 tw * 4);
   cairo_matrix_t matrix;
   cairo_get_matrix(cr, &matrix);
@@ -168,36 +170,38 @@ static void read_tile(openslide_t *osr,
 
 static void paint_region(openslide_t *osr, cairo_t *cr,
 			 int64_t x, int64_t y,
-			 int32_t layer, int32_t w, int32_t h) {
-  struct ngr_data *data = (struct ngr_data *) osr->data;
-  struct _openslide_ngr *ngr = data->ngrs[layer];
+			 int32_t level, int32_t w, int32_t h) {
+  struct ngr_data *data = osr->data;
+  struct _openslide_ngr *ngr = data->ngrs[level];
 
   // compute coordinates
-  double ds = openslide_get_layer_downsample(osr, layer);
+  double ds = openslide_get_level_downsample(osr, level);
   double ds_x = x / ds;
   double ds_y = y / ds;
   int64_t start_tile_x = ds_x / ngr->column_width;
-  int64_t end_tile_x = ((ds_x + w) / ngr->column_width) + 1;
-  int64_t start_tile_y = ds_y;
-  int64_t end_tile_y = ds_y + h + 1;
+  int64_t end_tile_x = ceil((ds_x + w) / ngr->column_width);
+  int64_t start_tile_y = ds_y / NGR_TILE_HEIGHT;
+  int64_t end_tile_y = ceil((ds_y + h) / NGR_TILE_HEIGHT);
 
   double offset_x = ds_x - (start_tile_x * ngr->column_width);
+  double offset_y = ds_y - (start_tile_y * NGR_TILE_HEIGHT);
 
   _openslide_read_tiles(cr,
-			layer,
+			level,
 			start_tile_x, start_tile_y,
 			end_tile_x, end_tile_y,
-			offset_x, 0,
-			ngr->column_width, 1,
-			osr, osr->cache,
+			offset_x, offset_y,
+			ngr->column_width, NGR_TILE_HEIGHT,
+			osr, osr->cache, NULL,
 			read_tile);
 }
 
 
 static const struct _openslide_ops _openslide_vmu_ops = {
-  get_dimensions,
-  paint_region,
-  destroy
+  .get_dimensions = get_dimensions,
+  .get_tile_geometry = get_tile_geometry,
+  .paint_region = paint_region,
+  .destroy = destroy,
 };
 
 
@@ -216,14 +220,11 @@ void _openslide_add_ngr_ops(openslide_t *osr,
   data->ngr_count = ngr_count;
   data->ngrs = ngrs;
 
-  // init cache lock
-  data->cache_mutex = g_mutex_new();
-
   // store vmu-specific data into osr
   g_assert(osr->data == NULL);
   osr->data = data;
 
   // general osr data
-  osr->layer_count = ngr_count;
+  osr->level_count = ngr_count;
   osr->ops = &_openslide_vmu_ops;
 }
