@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2010 Carnegie Mellon University
+ *  Copyright (c) 2007-2012 Carnegie Mellon University
  *  Copyright (c) 2011 Google, Inc.
  *  All rights reserved.
  *
@@ -27,24 +27,35 @@
 #include <glib.h>
 #include <tiffio.h>
 
+#include <stdio.h>
 #include <string.h>
+#include <math.h>
 #include <cairo.h>
 
 #include "openslide-tilehelper.h"
 #include "openslide-hash.h"
 
-struct _openslide_tiffopsdata {
-  TIFF *tiff;
+#define HANDLE_CACHE_MAX 32
 
-  GMutex *tiff_mutex;
+struct _openslide_tiffopsdata {
+  GQueue *handle_cache;
+  GMutex *handle_cache_mutex;
+  char *filename;
 
   int32_t overlap_count;
   int32_t *overlaps;
-  int32_t *layers;
+  int32_t *levels;
 
   _openslide_tiff_tilereader_fn tileread;
 };
 
+// not thread-safe, like libtiff
+struct tiff_file_handle {
+  char *filename;
+  int64_t offset;
+  int64_t size;
+};
+
 struct tiff_associated_image_ctx {
   tdir_t directory;
 };
@@ -62,13 +73,6 @@ struct tiff_associated_image_ctx {
   }								\
   result = tmp;
 
-static char *serialize_double(double d) {
-  char buf[G_ASCII_DTOSTR_BUF_SIZE];
-  g_ascii_dtostr (buf, sizeof buf, d);
-
-  return g_strdup(buf);
-}
-
 static const char *store_string_property(TIFF *tiff, GHashTable *ht,
 					 const char *name, ttag_t tag) {
   char *value;
@@ -91,7 +95,7 @@ static void store_float_property(TIFF *tiff, GHashTable *ht,
 				 const char *name, ttag_t tag) {
   float value;
   if (TIFFGetFieldDefaulted(tiff, tag, &value)) {
-    g_hash_table_insert(ht, g_strdup(name), serialize_double(value));
+    g_hash_table_insert(ht, g_strdup(name), _openslide_format_double(value));
   }
 }
 
@@ -151,36 +155,85 @@ static void store_and_hash_properties(TIFF *tiff, GHashTable *ht,
   }
 }
 
+static TIFF *get_tiff(openslide_t *osr) {
+  struct _openslide_tiffopsdata *data = osr->data;
+  TIFF *tiff;
+
+  //g_debug("get TIFF");
+  g_mutex_lock(data->handle_cache_mutex);
+  tiff = g_queue_pop_head(data->handle_cache);
+  g_mutex_unlock(data->handle_cache_mutex);
+
+  if (tiff == NULL) {
+    //g_debug("create TIFF");
+    // Does not check that we have the same file.  Then again, neither does
+    // tiff_do_read.
+    tiff = _openslide_tiff_open(data->filename);
+  }
+  if (tiff == NULL) {
+    _openslide_set_error(osr, "Cannot open TIFF file");
+  }
+  return tiff;
+}
+
+static void put_tiff(openslide_t *osr, TIFF *tiff) {
+  struct _openslide_tiffopsdata *data = osr->data;
+
+  if (tiff == NULL) {
+    return;
+  }
+
+  //g_debug("put TIFF");
+  g_mutex_lock(data->handle_cache_mutex);
+  if (g_queue_get_length(data->handle_cache) < HANDLE_CACHE_MAX) {
+    g_queue_push_head(data->handle_cache, tiff);
+    tiff = NULL;
+  }
+  g_mutex_unlock(data->handle_cache_mutex);
+
+  if (tiff) {
+    //g_debug("too many TIFFs");
+    TIFFClose(tiff);
+  }
+}
+
 static void destroy_data(struct _openslide_tiffopsdata *data) {
-  TIFFClose(data->tiff);
-  g_mutex_free(data->tiff_mutex);
-  g_free(data->layers);
+  TIFF *tiff;
+
+  g_mutex_lock(data->handle_cache_mutex);
+  while ((tiff = g_queue_pop_head(data->handle_cache)) != NULL) {
+    TIFFClose(tiff);
+  }
+  g_mutex_unlock(data->handle_cache_mutex);
+  g_queue_free(data->handle_cache);
+  g_mutex_free(data->handle_cache_mutex);
+  g_free(data->filename);
+  g_free(data->levels);
   g_free(data->overlaps);
   g_slice_free(struct _openslide_tiffopsdata, data);
 }
 
 static void destroy(openslide_t *osr) {
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
+  struct _openslide_tiffopsdata *data = osr->data;
   destroy_data(data);
 }
 
 
-static void get_dimensions_unlocked(openslide_t *osr, int32_t layer,
-				    int64_t *w, int64_t *h) {
+static void _get_dimensions(openslide_t *osr, TIFF *tiff,
+                            int32_t level, int64_t *w, int64_t *h) {
   uint32_t tmp;
 
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
-  TIFF *tiff = data->tiff;
+  struct _openslide_tiffopsdata *data = osr->data;
 
   int32_t ox = 0;
   int32_t oy = 0;
-  if (layer < data->overlap_count) {
-    ox = data->overlaps[layer * 2];
-    oy = data->overlaps[(layer * 2) + 1];
+  if (level < data->overlap_count) {
+    ox = data->overlaps[level * 2];
+    oy = data->overlaps[(level * 2) + 1];
   }
 
-  // set the layer
-  SET_DIR_OR_FAIL(osr, tiff, data->layers[layer])
+  // set the level
+  SET_DIR_OR_FAIL(osr, tiff, data->levels[level])
 
   // figure out tile size
   int64_t tw, th;
@@ -211,28 +264,65 @@ static void get_dimensions_unlocked(openslide_t *osr, int32_t layer,
   *h = ih_minus_o;
 }
 
-static void get_dimensions(openslide_t *osr, int32_t layer,
+static void get_dimensions(openslide_t *osr, int32_t level,
 			   int64_t *w, int64_t *h) {
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
+  TIFF *tiff = get_tiff(osr);
+  if (tiff != NULL) {
+    _get_dimensions(osr, tiff, level, w, h);
+  }
+  put_tiff(osr, tiff);
+}
 
-  g_mutex_lock(data->tiff_mutex);
-  get_dimensions_unlocked(osr, layer, w, h);
-  g_mutex_unlock(data->tiff_mutex);
+static void _get_tile_geometry(openslide_t *osr, TIFF *tiff, int32_t level,
+                               int64_t *w, int64_t *h) {
+  struct _openslide_tiffopsdata *data = osr->data;
+
+  uint32_t tmp;
+
+  // if any level has overlaps, reporting tile advances would mislead the
+  // application
+  for (int32_t i = 0; i < 2 * data->overlap_count; i++) {
+    if (data->overlaps[i]) {
+      return;
+    }
+  }
+
+  // set the level
+  SET_DIR_OR_FAIL(osr, tiff, data->levels[level])
+
+  // figure out tile size
+  int64_t tw, th;
+  GET_FIELD_OR_FAIL(osr, tiff, TIFFTAG_TILEWIDTH, tw)
+  GET_FIELD_OR_FAIL(osr, tiff, TIFFTAG_TILELENGTH, th)
+
+  // commit
+  *w = tw;
+  *h = th;
+}
+
+static void get_tile_geometry(openslide_t *osr, int32_t level,
+                              int64_t *w, int64_t *h) {
+  TIFF *tiff = get_tiff(osr);
+  if (tiff != NULL) {
+    _get_tile_geometry(osr, tiff, level, w, h);
+  }
+  put_tiff(osr, tiff);
 }
 
 static void read_tile(openslide_t *osr,
 		      cairo_t *cr,
-		      int32_t layer,
+		      int32_t level,
 		      int64_t tile_x, int64_t tile_y,
 		      double translate_x, double translate_y,
-		      struct _openslide_cache *cache) {
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
-  TIFF *tiff = data->tiff;
+		      struct _openslide_cache *cache,
+		      void *arg) {
+  struct _openslide_tiffopsdata *data = osr->data;
+  TIFF *tiff = arg;
 
   uint32_t tmp;
 
-  // set the layer
-  SET_DIR_OR_FAIL(osr, tiff, data->layers[layer])
+  // set the level
+  SET_DIR_OR_FAIL(osr, tiff, data->levels[level])
 
   // figure out tile size
   int64_t tw, th;
@@ -253,14 +343,10 @@ static void read_tile(openslide_t *osr,
 
   // cache
   struct _openslide_cache_entry *cache_entry;
-  uint32_t *tiledata = (uint32_t *) _openslide_cache_get(cache,
-							 x,
-							 y,
-							 layer,
-							 &cache_entry);
+  uint32_t *tiledata = _openslide_cache_get(cache, x, y, level, &cache_entry);
   if (!tiledata) {
-    tiledata = (uint32_t *) g_slice_alloc(tw * th * 4);
-    data->tileread(osr, data->tiff, tiledata, x, y, tw, th);
+    tiledata = g_slice_alloc(tw * th * 4);
+    data->tileread(osr, tiff, tiledata, x, y, tw, th);
 
     // clip, if necessary
     int64_t rx = iw - x;
@@ -286,7 +372,7 @@ static void read_tile(openslide_t *osr,
     }
 
     // put it in the cache
-    _openslide_cache_put(cache, x, y, layer,
+    _openslide_cache_put(cache, x, y, level,
 			 tiledata, tw * th * 4,
 			 &cache_entry);
   }
@@ -331,16 +417,15 @@ static void read_tile(openslide_t *osr,
   _openslide_cache_entry_unref(cache_entry);
 }
 
-static void paint_region_unlocked(openslide_t *osr, cairo_t *cr,
-				  int64_t x, int64_t y,
-				  int32_t layer,
-				  int32_t w, int32_t h) {
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
-  TIFF *tiff = data->tiff;
+static void _paint_region(openslide_t *osr, TIFF *tiff, cairo_t *cr,
+                          int64_t x, int64_t y,
+                          int32_t level,
+                          int32_t w, int32_t h) {
+  struct _openslide_tiffopsdata *data = osr->data;
   uint32_t tmp;
 
-  // set the layer
-  SET_DIR_OR_FAIL(osr, tiff, data->layers[layer])
+  // set the level
+  SET_DIR_OR_FAIL(osr, tiff, data->levels[level])
 
   // figure out tile size
   int64_t tw, th;
@@ -359,18 +444,18 @@ static void paint_region_unlocked(openslide_t *osr, cairo_t *cr,
   // compute coordinates
   int32_t ox = 0;
   int32_t oy = 0;
-  if (layer < data->overlap_count) {
-    ox = data->overlaps[layer * 2];
-    oy = data->overlaps[(layer * 2) + 1];
+  if (level < data->overlap_count) {
+    ox = data->overlaps[level * 2];
+    oy = data->overlaps[(level * 2) + 1];
   }
 
-  double ds = openslide_get_layer_downsample(osr, layer);
+  double ds = openslide_get_level_downsample(osr, level);
   double ds_x = x / ds;
   double ds_y = y / ds;
   int64_t start_tile_x = ds_x / (tw - ox);
-  int64_t end_tile_x = ((ds_x + w) / (tw - ox)) + 1;
+  int64_t end_tile_x = ceil((ds_x + w) / (tw - ox));
   int64_t start_tile_y = ds_y / (th - oy);
-  int64_t end_tile_y = ((ds_y + h) / (th - oy)) + 1;
+  int64_t end_tile_y = ceil((ds_y + h) / (th - oy));
 
   double offset_x = ds_x - (start_tile_x * (tw - ox));
   double offset_y = ds_y - (start_tile_y * (th - oy));
@@ -401,78 +486,88 @@ static void paint_region_unlocked(openslide_t *osr, cairo_t *cr,
     }
   }
 
-  _openslide_read_tiles(cr, layer,
+  _openslide_read_tiles(cr, level,
 			start_tile_x, start_tile_y,
 			end_tile_x, end_tile_y,
 			offset_x, offset_y,
 			advance_x, advance_y,
-			osr, osr->cache,
+			osr, osr->cache, tiff,
 			read_tile);
 }
 
 static void paint_region(openslide_t *osr, cairo_t *cr,
 			 int64_t x, int64_t y,
-			 int32_t layer,
+			 int32_t level,
 			 int32_t w, int32_t h) {
-  struct _openslide_tiffopsdata *data = (struct _openslide_tiffopsdata *) osr->data;
-
-  g_mutex_lock(data->tiff_mutex);
-  paint_region_unlocked(osr, cr, x, y, layer, w, h);
-  g_mutex_unlock(data->tiff_mutex);
+  TIFF *tiff = get_tiff(osr);
+  if (tiff) {
+    _paint_region(osr, tiff, cr, x, y, level, w, h);
+  }
+  put_tiff(osr, tiff);
 }
 
 
 static const struct _openslide_ops _openslide_tiff_ops = {
-  get_dimensions,
-  paint_region,
-  destroy
+  .get_dimensions = get_dimensions,
+  .get_tile_geometry = get_tile_geometry,
+  .paint_region = paint_region,
+  .destroy = destroy,
 };
 
 void _openslide_add_tiff_ops(openslide_t *osr,
 			     TIFF *tiff,
+			     int32_t property_dir,
 			     int32_t overlap_count,
 			     int32_t *overlaps,
-			     int32_t layer_count,
-			     int32_t *layers,
+			     int32_t level_count,
+			     int32_t *levels,
 			     _openslide_tiff_tilereader_fn tileread,
 			     struct _openslide_hash *quickhash1) {
   // allocate private data
   struct _openslide_tiffopsdata *data =
     g_slice_new(struct _openslide_tiffopsdata);
 
-  // store layer info
-  data->layers = layers;
+  GError *tmp_err = NULL;
+
+  // store level info
+  data->levels = levels;
 
   // populate private data
-  data->tiff = tiff;
-  data->tiff_mutex = g_mutex_new();
+  data->handle_cache = g_queue_new();
+  data->handle_cache_mutex = g_mutex_new();
+  data->filename = g_strdup(TIFFFileName(tiff));
   data->tileread = tileread;
   data->overlap_count = overlap_count;
   data->overlaps = overlaps;
 
   if (osr == NULL) {
     // free now and return
+    TIFFClose(tiff);
     destroy_data(data);
     return;
   }
 
-  // generate hash of the smallest layer
-  TIFFSetDirectory(data->tiff, layers[layer_count - 1]);
-  if (!_openslide_hash_tiff_tiles(quickhash1, tiff)) {
-    _openslide_set_error(osr, "Cannot hash TIFF tiles");
+  // generate hash of the smallest level
+  TIFFSetDirectory(tiff, levels[level_count - 1]);
+  if (!_openslide_hash_tiff_tiles(quickhash1, tiff, &tmp_err)) {
+    _openslide_set_error(osr, "Cannot hash TIFF tiles: %s", tmp_err->message);
+    g_clear_error(&tmp_err);
   }
 
   // load TIFF properties
-  TIFFSetDirectory(data->tiff, 0);    // ignoring return value, but nothing we can do if failed
-  store_and_hash_properties(data->tiff, osr->properties, quickhash1);
+  TIFFSetDirectory(tiff, property_dir);    // ignoring return value, but nothing we can do if failed
+  store_and_hash_properties(tiff, osr->properties, quickhash1);
 
   // store tiff-specific data into osr
   g_assert(osr->data == NULL);
 
   // general osr data
-  osr->layer_count = layer_count;
+  osr->level_count = level_count;
   osr->data = data;
   osr->ops = &_openslide_tiff_ops;
+
+  // now store TIFF handle
+  put_tiff(osr, tiff);
 }
 
 void _openslide_generic_tiff_tilereader(openslide_t *osr,
@@ -497,7 +592,7 @@ void _openslide_generic_tiff_tilereader(openslide_t *osr,
   img.row_offset = y;
 
   // draw it
-  if (TIFFRGBAImageGet(&img, (uint32*)dest, w, h)) {
+  if (TIFFRGBAImageGet(&img, dest, w, h)) {
     // permute
     uint32_t *p = dest;
     uint32_t *end = dest + w * h;
@@ -516,15 +611,11 @@ void _openslide_generic_tiff_tilereader(openslide_t *osr,
   TIFFRGBAImageEnd(&img);
 }
 
-static void tiff_get_associated_image_data_unlocked(openslide_t *osr,
-                                                    void *_ctx,
-                                                    uint32_t *dest,
-                                                    int64_t w, int64_t h) {
-  struct _openslide_tiffopsdata *data =
-    (struct _openslide_tiffopsdata *) osr->data;
-  struct tiff_associated_image_ctx *ctx =
-    (struct tiff_associated_image_ctx *) _ctx;
-  TIFF *tiff = data->tiff;
+static void _tiff_get_associated_image_data(openslide_t *osr, TIFF *tiff,
+                                            void *_ctx,
+                                            uint32_t *dest,
+                                            int64_t w, int64_t h) {
+  struct tiff_associated_image_ctx *ctx = _ctx;
   uint32_t tmp;
   int64_t width, height;
 
@@ -548,35 +639,36 @@ static void tiff_get_associated_image_data(openslide_t *osr, void *ctx,
                                            uint32_t *dest,
                                            int64_t w, int64_t h)
 {
-  struct _openslide_tiffopsdata *data =
-    (struct _openslide_tiffopsdata *) osr->data;
-
-  g_mutex_lock(data->tiff_mutex);
-  tiff_get_associated_image_data_unlocked(osr, ctx, dest, w, h);
-  g_mutex_unlock(data->tiff_mutex);
+  TIFF *tiff = get_tiff(osr);
+  if (tiff) {
+    _tiff_get_associated_image_data(osr, tiff, ctx, dest, w, h);
+  }
+  put_tiff(osr, tiff);
 }
 
 static void tiff_destroy_associated_image_ctx(void *_ctx) {
-  struct tiff_associated_image_ctx *ctx =
-    (struct tiff_associated_image_ctx *) _ctx;
+  struct tiff_associated_image_ctx *ctx = _ctx;
 
   g_slice_free(struct tiff_associated_image_ctx, ctx);
 }
 
 bool _openslide_add_tiff_associated_image(GHashTable *ht,
 					  const char *name,
-					  TIFF *tiff) {
+					  TIFF *tiff,
+					  GError **err) {
   uint32_t tmp;
 
   // get the dimensions
   if (!TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &tmp)) {
-    g_warning("Cannot get associated image width");
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_BAD_DATA,
+                "Cannot get associated image width");
     return false;
   }
   int64_t w = tmp;
 
   if (!TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &tmp)) {
-    g_warning("Cannot get associated image height");
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_BAD_DATA,
+                "Cannot get associated image height");
     return false;
   }
   int64_t h = tmp;
@@ -601,3 +693,127 @@ bool _openslide_add_tiff_associated_image(GHashTable *ht,
 
   return true;
 }
+
+static tsize_t tiff_do_read(thandle_t th, tdata_t buf, tsize_t size) {
+  struct tiff_file_handle *hdl = th;
+
+  // don't leave the file handle open between calls
+  // also ensures FD_CLOEXEC is set
+  FILE *f = _openslide_fopen(hdl->filename, "rb", NULL);
+  if (f == NULL) {
+    return 0;
+  }
+  if (fseeko(f, hdl->offset, SEEK_SET)) {
+    fclose(f);
+    return 0;
+  }
+  int64_t rsize = fread(buf, 1, size, f);
+  hdl->offset += rsize;
+  fclose(f);
+  return rsize;
+}
+
+static tsize_t tiff_do_write(thandle_t th G_GNUC_UNUSED,
+                             tdata_t data G_GNUC_UNUSED,
+                             tsize_t size G_GNUC_UNUSED) {
+  // fail
+  return 0;
+}
+
+static toff_t tiff_do_seek(thandle_t th, toff_t offset, int whence) {
+  struct tiff_file_handle *hdl = th;
+
+  switch (whence) {
+  case SEEK_SET:
+    hdl->offset = offset;
+    break;
+  case SEEK_CUR:
+    hdl->offset += offset;
+    break;
+  case SEEK_END:
+    hdl->offset = hdl->size + offset;
+    break;
+  default:
+    g_assert_not_reached();
+  }
+  return hdl->offset;
+}
+
+static int tiff_do_close(thandle_t th) {
+  struct tiff_file_handle *hdl = th;
+
+  g_free(hdl->filename);
+  g_slice_free(struct tiff_file_handle, hdl);
+  return 0;
+}
+
+static toff_t tiff_do_size(thandle_t th) {
+  struct tiff_file_handle *hdl = th;
+
+  return hdl->size;
+}
+
+TIFF *_openslide_tiff_open(const char *filename) {
+  // open
+  FILE *f = _openslide_fopen(filename, "rb", NULL);
+  if (f == NULL) {
+    return NULL;
+  }
+
+  // read magic
+  uint8_t buf[4];
+  if (fread(buf, 4, 1, f) != 1) {
+    // can't read
+    fclose(f);
+    return NULL;
+  }
+
+  // get size
+  if (fseeko(f, 0, SEEK_END) == -1) {
+    fclose(f);
+    return NULL;
+  }
+  int64_t size = ftello(f);
+  fclose(f);
+  if (size == -1) {
+    return NULL;
+  }
+
+  // check magic
+  // TODO: remove if libtiff gets private error/warning callbacks
+  if (buf[0] != buf[1]) {
+    return NULL;
+  }
+  uint16_t version;
+  switch (buf[0]) {
+  case 'M':
+    // big endian
+    version = (buf[2] << 8) | buf[3];
+    break;
+  case 'I':
+    // little endian
+    version = (buf[3] << 8) | buf[2];
+    break;
+  default:
+    return NULL;
+  }
+  // only accept BigTIFF on libtiff >= 4
+  if (!(version == 42 || (sizeof(toff_t) > 4 && version == 43))) {
+    return NULL;
+  }
+
+  // allocate
+  struct tiff_file_handle *hdl = g_slice_new0(struct tiff_file_handle);
+  hdl->filename = g_strdup(filename);
+  hdl->size = size;
+
+  // TIFFOpen
+  // mode: m disables mmap to avoid sigbus and other mmap fragility
+  TIFF *tiff = TIFFClientOpen(filename, "rm", hdl,
+                              tiff_do_read, tiff_do_write, tiff_do_seek,
+                              tiff_do_close, tiff_do_size, NULL, NULL);
+  if (tiff == NULL) {
+    tiff_do_close(hdl);
+  }
+  return tiff;
+}
