@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2010 Carnegie Mellon University
+ *  Copyright (c) 2007-2012 Carnegie Mellon University
  *  All rights reserved.
  *
  *  OpenSlide is free software: you can redistribute it and/or modify
@@ -24,12 +24,6 @@
 
 #ifdef _WIN32
 #define WIN32 1
-#define __MSVCRT_VERSION__ 0x0800
-#endif
-
-#if !defined(fseeko) && defined(_WIN32)
-#define fseeko(stream, offset, origin) _fseeki64(stream, offset, origin)
-#define ftello(stream) _ftelli64(stream)
 #endif
 
 #include <config.h>
@@ -38,27 +32,21 @@
 #include "openslide-hash.h"
 
 #include <glib.h>
-#ifndef _MSC_VER
 #include <stdbool.h>
-#endif
 #include <setjmp.h>
 #include <tiffio.h>
 
-
-// TODO detect if libjpeg is built as a C++ library?
-#ifdef __cplusplus
-extern "C" {
-#endif
+// jconfig.h redefines HAVE_STDLIB_H if libjpeg was not built with Autoconf
+#undef HAVE_STDLIB_H
 #include <jpeglib.h>
-#ifdef __cplusplus
-};
-#endif
-
+#undef HAVE_STDLIB_H
+#include <config.h>  // again
 
 #include <cairo.h>
 
 #include <openjpeg.h>
 
+
 /* the associated image structure */
 struct _openslide_associated_image {
   int64_t w;
@@ -75,7 +63,7 @@ struct _openslide_associated_image {
 struct _openslide {
   const struct _openslide_ops *ops;
   void *data;
-  int32_t layer_count;
+  int32_t level_count;
 
   double *downsamples;  // zero values or NULL are filled in automatically from dimensions
 
@@ -97,20 +85,26 @@ struct _openslide {
 /* the function pointer structure for backends */
 struct _openslide_ops {
   void (*get_dimensions)(openslide_t *osr,
-			 int32_t layer,
+			 int32_t level,
 			 int64_t *w, int64_t *h);
+  // need not do anything, but must be consistent
+  void (*get_tile_geometry)(openslide_t *osr,
+			    int32_t level,
+			    int64_t *w, int64_t *h);
   void (*paint_region)(openslide_t *osr, cairo_t *cr,
 		       int64_t x, int64_t y,
-		       int32_t layer,
+		       int32_t level,
 		       int32_t w, int32_t h);
   void (*destroy)(openslide_t *osr);
 };
 
 /* vendor detection and parsing */
 typedef bool (*_openslide_vendor_fn)(openslide_t *osr, const char *filename,
-				     struct _openslide_hash *quickhash1);
+				     struct _openslide_hash *quickhash1,
+				     GError **err);
 typedef bool (*_openslide_tiff_vendor_fn)(openslide_t *osr, TIFF *tiff,
-					  struct _openslide_hash *quickhash1);
+					  struct _openslide_hash *quickhash1,
+					  GError **err);
 /*
  * A note on quickhash1: this should be a hash of data that
  * will not change with revisions to the openslide library. It should
@@ -128,17 +122,23 @@ typedef bool (*_openslide_tiff_vendor_fn)(openslide_t *osr, TIFF *tiff,
 
 
 bool _openslide_try_trestle(openslide_t *osr, TIFF *tiff,
-			    struct _openslide_hash *quickhash1);
+			    struct _openslide_hash *quickhash1, GError **err);
 bool _openslide_try_aperio(openslide_t *osr, TIFF *tiff,
-			   struct _openslide_hash *quickhash1);
+			   struct _openslide_hash *quickhash1, GError **err);
 bool _openslide_try_hamamatsu(openslide_t *osr, const char* filename,
-			      struct _openslide_hash *quickhash1);
+			      struct _openslide_hash *quickhash1,
+			      GError **err);
 bool _openslide_try_hamamatsu_ndpi(openslide_t *osr, const char* filename,
-				   struct _openslide_hash *quickhash1);
+				   struct _openslide_hash *quickhash1,
+				   GError **err);
 bool _openslide_try_mirax(openslide_t *osr, const char* filename,
-			  struct _openslide_hash *quickhash1);
+			  struct _openslide_hash *quickhash1, GError **err);
+bool _openslide_try_leica(openslide_t *osr, TIFF *tiff,
+				 struct _openslide_hash *quickhash1,
+				 GError **err);
 bool _openslide_try_generic_tiff(openslide_t *osr, TIFF *tiff,
-				 struct _openslide_hash *quickhash1);
+				 struct _openslide_hash *quickhash1,
+				 GError **err);
 
 
 /* GHashTable utils */
@@ -148,10 +148,27 @@ void _openslide_int64_free(gpointer data);
 
 /* g_key_file_load_from_file wrapper */
 gboolean _openslide_read_key_file(GKeyFile *key_file, const char *filename,
-                                  GKeyFileFlags flags, GError **error);
+                                  GKeyFileFlags flags, GError **err);
 
 /* fopen() wrapper which properly sets FD_CLOEXEC */
-FILE *_openslide_fopen(const char *path, const char *mode);
+FILE *_openslide_fopen(const char *path, const char *mode, GError **err);
+
+/* Returns the size of the file */
+int64_t _openslide_fsize(const char *path, GError **err);
+
+/* Serialize double to string */
+char *_openslide_format_double(double d);
+
+/* Duplicate OpenSlide properties */
+void _openslide_duplicate_int_prop(GHashTable *ht, const char *src,
+                                   const char *dest);
+void _openslide_duplicate_double_prop(GHashTable *ht, const char *src,
+                                      const char *dest);
+
+// background color helper
+void _openslide_set_background_color_prop(GHashTable *ht,
+                                          uint8_t r, uint8_t g, uint8_t b);
+
 
 /* TIFF support */
 typedef void (*_openslide_tiff_tilereader_fn)(openslide_t *osr,
@@ -164,10 +181,11 @@ typedef void (*_openslide_tiff_tilereader_fn)(openslide_t *osr,
 
 void _openslide_add_tiff_ops(openslide_t *osr,
 			     TIFF *tiff,
+			     int32_t property_dir,
 			     int32_t overlap_count,
 			     int32_t *overlaps,
-			     int32_t layer_count,
-			     int32_t *layers,
+			     int32_t level_count,
+			     int32_t *levels,
 			     _openslide_tiff_tilereader_fn tileread,
 			     struct _openslide_hash *quickhash1);
 
@@ -179,7 +197,10 @@ void _openslide_generic_tiff_tilereader(openslide_t *osr,
 
 bool _openslide_add_tiff_associated_image(GHashTable *ht,
 					  const char *name,
-					  TIFF *tiff);
+					  TIFF *tiff,
+					  GError **err);
+
+TIFF *_openslide_tiff_open(const char *filename);
 
 /* JPEG support */
 struct _openslide_jpeg_file {
@@ -213,12 +234,12 @@ struct _openslide_jpeg_tile {
   double dest_offset_y;
 };
 
-struct _openslide_jpeg_layer {
+struct _openslide_jpeg_level {
   GHashTable *tiles;
 
   // size of canvas
-  int64_t layer_w;
-  int64_t layer_h;
+  int64_t level_w;
+  int64_t level_h;
 
   int32_t tiles_across;
   int32_t tiles_down;
@@ -238,15 +259,16 @@ struct _openslide_jpeg_layer {
 void _openslide_add_jpeg_ops(openslide_t *osr,
 			     int32_t file_count,
 			     struct _openslide_jpeg_file **files,
-			     int32_t layer_count,
-			     struct _openslide_jpeg_layer **layers);
+			     int32_t level_count,
+			     struct _openslide_jpeg_level **levels);
 
 GHashTable *_openslide_jpeg_create_tiles_table(void);
 
 bool _openslide_add_jpeg_associated_image(GHashTable *ht,
 					  const char *name,
 					  const char *filename,
-					  int64_t offset);
+					  int64_t offset,
+					  GError **err);
 
 /*
  * On Windows, we cannot fopen a file and pass it to another DLL that does fread.
@@ -259,9 +281,10 @@ struct _openslide_jpeg_error_mgr {
   struct jpeg_error_mgr pub;      // public fields
 
   jmp_buf *env;
+  GError *err;
 };
 
-struct jpeg_error_mgr *_openslide_jpeg_set_error_handler(struct _openslide_jpeg_error_mgr *err,
+struct jpeg_error_mgr *_openslide_jpeg_set_error_handler(struct _openslide_jpeg_error_mgr *jerr,
 							 jmp_buf *env);
 
 // Hamamatsu NGR
@@ -281,21 +304,31 @@ void _openslide_add_ngr_ops(openslide_t *osr,
 			    struct _openslide_ngr **ngrs);
 
 
-// error handling
+// external error propagation
 bool _openslide_set_error(openslide_t *osr, const char *format, ...);
 bool _openslide_check_cairo_status_possibly_set_error(openslide_t *osr,
 						      cairo_t *cr);
 
+// internal error propagation
+enum OpenSlideError {
+  // file format unrecognized; try other formats
+  OPENSLIDE_ERROR_FORMAT_NOT_SUPPORTED,
+  // file corrupt; hard fail
+  OPENSLIDE_ERROR_BAD_DATA,
+};
+#define OPENSLIDE_ERROR _openslide_error_quark()
+GQuark _openslide_error_quark(void);
 
-// background color helper
-void _openslide_set_background_color_property(GHashTable *ht,
-					      uint8_t r, uint8_t g, uint8_t b);
+void _openslide_io_error(GError **err, const char *fmt, ...);
+void _openslide_set_error_from_gerror(openslide_t *osr, GError *err);
 
 // private properties, for now
-#define _OPENSLIDE_PROPERTY_NAME_LAYER_COUNT "openslide.layer-count"
-#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_WIDTH "openslide.layer[%d].width"
-#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_HEIGHT "openslide.layer[%d].height"
-#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_DOWNSAMPLE "openslide.layer[%d].downsample"
+#define _OPENSLIDE_PROPERTY_NAME_LEVEL_COUNT "openslide.level-count"
+#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_WIDTH "openslide.level[%d].width"
+#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_HEIGHT "openslide.level[%d].height"
+#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_DOWNSAMPLE "openslide.level[%d].downsample"
+#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_TILE_WIDTH "openslide.level[%d].tile-width"
+#define _OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_TILE_HEIGHT "openslide.level[%d].tile-height"
 
 // deprecated prefetch stuff (maybe we'll undeprecate it someday),
 // still needs these declarations for ABI compat
@@ -304,7 +337,7 @@ void _openslide_set_background_color_property(GHashTable *ht,
 OPENSLIDE_PUBLIC()
 int openslide_give_prefetch_hint(openslide_t *osr,
 				 int64_t x, int64_t y,
-				 int32_t layer,
+				 int32_t level,
 				 int64_t w, int64_t h);
 #undef openslide_cancel_prefetch_hint
 OPENSLIDE_PUBLIC()
