@@ -27,11 +27,14 @@
 #include <string.h>
 
 #include <glib.h>
+#include <libxml/parser.h>
 
 #include "openslide-cache.h"
 #include "openslide-tilehelper.h"
 #include "openslide-cairo.h"
 
+const char _openslide_release_info[] = "OpenSlide " PACKAGE_VERSION ", copyright (C) 2007-2012 Carnegie Mellon University and others.\nLicensed under the GNU Lesser General Public License, version 2.1.";
+
 static const char * const EMPTY_STRING_ARRAY[] = { NULL };
 
 static const _openslide_vendor_fn non_tiff_formats[] = {
@@ -44,32 +47,23 @@ static const _openslide_vendor_fn non_tiff_formats[] = {
 static const _openslide_tiff_vendor_fn tiff_formats[] = {
   _openslide_try_trestle,
   _openslide_try_aperio,
+  _openslide_try_leica,
   _openslide_try_generic_tiff,
   NULL
 };
 
 static bool openslide_was_dynamically_loaded;
 
-#ifdef _MSC_VER
-  #pragma section(".CRT$XCU",read)
-  #define INITIALIZER(f) \
-  static void __cdecl f(void); \
-  __declspec(allocate(".CRT$XCU")) void (__cdecl*f##_)(void) = f; \
-  static void __cdecl f(void)
-#elif defined(__GNUC__)
-  #define INITIALIZER(f) \
-  static void __attribute__((constructor)) f(void)
-#endif
-
 // called from shared-library constructor!
-INITIALIZER(_openslide_init) {
+static void __attribute__((constructor)) _openslide_init(void) {
   // activate threads
   if (!g_thread_supported ()) g_thread_init (NULL);
+  xmlInitParser();
   openslide_was_dynamically_loaded = true;
 }
 
 static void destroy_associated_image(gpointer data) {
-  struct _openslide_associated_image *img = (struct _openslide_associated_image *) data;
+  struct _openslide_associated_image *img = data;
 
   if (img->destroy_ctx != NULL && img->ctx != NULL) {
     img->destroy_ctx(img->ctx);
@@ -77,54 +71,18 @@ static void destroy_associated_image(gpointer data) {
   g_slice_free(struct _openslide_associated_image, img);
 }
 
-static bool layer_in_range(openslide_t *osr, int32_t layer) {
-  if (layer < 0) {
+static bool level_in_range(openslide_t *osr, int32_t level) {
+  if (level < 0) {
     return false;
   }
 
-  if (layer > osr->layer_count - 1) {
+  if (level > osr->level_count - 1) {
     return false;
   }
 
   return true;
 }
 
-// TODO: remove entirely if libtiff gets private error/warning callbacks
-static bool quick_tiff_check(const char *filename) {
-  FILE *f = _openslide_fopen(filename, "rb");
-  if (f == NULL) {
-    return false;
-  }
-
-  // read magic
-  uint8_t buf[4];
-  if (fread(buf, 4, 1, f) != 1) {
-    // can't read
-    fclose(f);
-    return false;
-  }
-
-  fclose(f);
-
-  // check magic
-  if (buf[0] != buf[1]) {
-    return false;
-  }
-
-  switch (buf[0]) {
-  case 'M':
-    // big endian
-    return (buf[2] == 0) && ((buf[3] == 42) || (buf[3] == 43));
-
-  case 'I':
-    // little endian
-    return (buf[3] == 0) && ((buf[2] == 42) || (buf[2] == 43));
-
-  default:
-    return false;
-  }
-}
-
 static void reset_osr(openslide_t *osr) {
   if (osr) {
     g_hash_table_remove_all(osr->properties);
@@ -146,54 +104,95 @@ static void free_quickhash1_if_failed(bool result,
   }
 }
 
+static void fixup_format_error(bool result, GError **err) {
+  // check for error-handling bugs in detector
+
+  if (!result && err && !*err) {
+    g_warning("Format detector failed without setting error");
+    // assume the worst
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_BAD_DATA,
+                "Unknown error");
+  }
+
+  if (result && err && *err) {
+    g_warning("Format detector succeeded but set error: %s", (*err)->message);
+    g_clear_error(err);
+  }
+}
+
 static bool try_format(openslide_t *osr, const char *filename,
 		       struct _openslide_hash **quickhash1_OUT,
-		       const _openslide_vendor_fn *format_check) {
+		       const _openslide_vendor_fn *format_check,
+		       GError **err) {
   reset_osr(osr);
   init_quickhash1_out(quickhash1_OUT);
 
-  bool result = (*format_check)(osr, filename, quickhash1_OUT ? *quickhash1_OUT : NULL);
+  bool result = (*format_check)(osr, filename,
+                                quickhash1_OUT ? *quickhash1_OUT : NULL,
+                                err);
 
   free_quickhash1_if_failed(result, quickhash1_OUT);
+  fixup_format_error(result, err);
 
   return result;
 }
 
 static bool try_tiff_format(openslide_t *osr, TIFF *tiff,
 			    struct _openslide_hash **quickhash1_OUT,
-			    const _openslide_tiff_vendor_fn *format_check) {
+			    const _openslide_tiff_vendor_fn *format_check,
+			    GError **err) {
   reset_osr(osr);
   init_quickhash1_out(quickhash1_OUT);
 
   TIFFSetDirectory(tiff, 0);
-  bool result = (*format_check)(osr, tiff, quickhash1_OUT ? *quickhash1_OUT : NULL);
+  bool result = (*format_check)(osr, tiff,
+                                quickhash1_OUT ? *quickhash1_OUT : NULL,
+                                err);
 
   free_quickhash1_if_failed(result, quickhash1_OUT);
+  fixup_format_error(result, err);
 
   return result;
 }
 
 static bool try_all_formats(openslide_t *osr, const char *filename,
-			    struct _openslide_hash **quickhash1_OUT) {
+			    struct _openslide_hash **quickhash1_OUT,
+			    GError **err) {
+  GError *tmp_err = NULL;
+
   // non-tiff
   const _openslide_vendor_fn *fn = non_tiff_formats;
   while (*fn) {
-    if (try_format(osr, filename, quickhash1_OUT, fn)) {
+    if (try_format(osr, filename, quickhash1_OUT, fn, &tmp_err)) {
       return true;
     }
+    if (!g_error_matches(tmp_err, OPENSLIDE_ERROR,
+                         OPENSLIDE_ERROR_FORMAT_NOT_SUPPORTED)) {
+      g_propagate_error(err, tmp_err);
+      return false;
+    }
+    //g_debug("%s", tmp_err->message);
+    g_clear_error(&tmp_err);
     fn++;
   }
 
 
   // tiff
-  TIFF *tiff;
-  // TIFFOpen: m disables mmap to avoid sigbus and other mmap fragility
-  if (quick_tiff_check(filename) && ((tiff = TIFFOpen(filename, "rm")) != NULL)) {
+  TIFF *tiff = _openslide_tiff_open(filename);
+  if (tiff != NULL) {
     const _openslide_tiff_vendor_fn *tfn = tiff_formats;
     while (*tfn) {
-      if (try_tiff_format(osr, tiff, quickhash1_OUT, tfn)) {
+      if (try_tiff_format(osr, tiff, quickhash1_OUT, tfn, &tmp_err)) {
 	return true;
       }
+      if (!g_error_matches(tmp_err, OPENSLIDE_ERROR,
+                           OPENSLIDE_ERROR_FORMAT_NOT_SUPPORTED)) {
+        g_propagate_error(err, tmp_err);
+        TIFFClose(tiff);
+        return false;
+      }
+      //g_debug("%s", tmp_err->message);
+      g_clear_error(&tmp_err);
       tfn++;
     }
 
@@ -203,6 +202,8 @@ static bool try_all_formats(openslide_t *osr, const char *filename,
 
 
   // no match
+  g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FORMAT_NOT_SUPPORTED,
+              "Format not recognized");
   return false;
 }
 
@@ -210,7 +211,7 @@ bool openslide_can_open(const char *filename) {
   g_assert(openslide_was_dynamically_loaded);
 
   // quick test
-  return try_all_formats(NULL, filename, NULL);
+  return try_all_formats(NULL, filename, NULL, NULL);
 }
 
 
@@ -222,9 +223,9 @@ struct add_key_to_strv_data {
 static void add_key_to_strv(gpointer key,
 			    gpointer value G_GNUC_UNUSED,
 			    gpointer user_data) {
-  struct add_key_to_strv_data *d = (struct add_key_to_strv_data *) user_data;
+  struct add_key_to_strv_data *d = user_data;
 
-  d->strv[d->i++] = (const char *) key;
+  d->strv[d->i++] = key;
 }
 
 static int cmpstring(const void *p1, const void *p2) {
@@ -244,6 +245,8 @@ static const char **strv_from_hashtable_keys(GHashTable *h) {
 }
 
 openslide_t *openslide_open(const char *filename) {
+  GError *tmp_err = NULL;
+
   g_assert(openslide_was_dynamically_loaded);
 
   // alloc memory
@@ -255,26 +258,36 @@ openslide_t *openslide_open(const char *filename) {
 
   // try to read it
   struct _openslide_hash *quickhash1 = NULL;
-  if (!try_all_formats(osr, filename, &quickhash1)) {
+  if (!try_all_formats(osr, filename, &quickhash1, &tmp_err)) {
     // failure
-    openslide_close(osr);
-    return NULL;
+    if (g_error_matches(tmp_err, OPENSLIDE_ERROR,
+                        OPENSLIDE_ERROR_FORMAT_NOT_SUPPORTED)) {
+      // not a slide file
+      g_clear_error(&tmp_err);
+      openslide_close(osr);
+      return NULL;
+    } else {
+      // failed to read slide
+      _openslide_set_error(osr, "%s", tmp_err->message);
+      g_clear_error(&tmp_err);
+      return osr;
+    }
   }
 
   // compute downsamples if not done already
   int64_t blw, blh;
-  openslide_get_layer0_dimensions(osr, &blw, &blh);
+  openslide_get_level0_dimensions(osr, &blw, &blh);
 
   if (!osr->downsamples) {
-    osr->downsamples = g_new0(double, osr->layer_count);
+    osr->downsamples = g_new0(double, osr->level_count);
   }
   if (osr->downsamples[0] == 0) {
     osr->downsamples[0] = 1.0;
   }
-  for (int32_t i = 1; i < osr->layer_count; i++) {
+  for (int32_t i = 1; i < osr->level_count; i++) {
     if (osr->downsamples[i] == 0) {
       int64_t w, h;
-      openslide_get_layer_dimensions(osr, i, &w, &h);
+      openslide_get_level_dimensions(osr, i, &w, &h);
 
       osr->downsamples[i] =
         (((double) blh / (double) h) +
@@ -283,7 +296,7 @@ openslide_t *openslide_open(const char *filename) {
   }
 
   // check downsamples
-  for (int32_t i = 1; i < osr->layer_count; i++) {
+  for (int32_t i = 1; i < osr->level_count; i++) {
     //g_debug("downsample: %g", osr->downsamples[i]);
 
     if (osr->downsamples[i] < osr->downsamples[i - 1]) {
@@ -306,24 +319,41 @@ openslide_t *openslide_open(const char *filename) {
 
   // set other properties
   g_hash_table_insert(osr->properties,
-		      g_strdup(_OPENSLIDE_PROPERTY_NAME_LAYER_COUNT),
-		      g_strdup_printf("%d", osr->layer_count));
-  for (int32_t i = 0; i < osr->layer_count; i++) {
+		      g_strdup(_OPENSLIDE_PROPERTY_NAME_LEVEL_COUNT),
+		      g_strdup_printf("%d", osr->level_count));
+  bool should_have_geometry;
+  for (int32_t i = 0; i < osr->level_count; i++) {
     int64_t w, h;
-    openslide_get_layer_dimensions(osr, i, &w, &h);
-
-    char downsample[G_ASCII_DTOSTR_BUF_SIZE];
-    g_ascii_dtostr(downsample, sizeof(downsample), osr->downsamples[i]);
+    openslide_get_level_dimensions(osr, i, &w, &h);
 
     g_hash_table_insert(osr->properties,
-			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_WIDTH, i),
+			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_WIDTH, i),
 			g_strdup_printf("%" G_GINT64_FORMAT, w));
     g_hash_table_insert(osr->properties,
-			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_HEIGHT, i),
+			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_HEIGHT, i),
 			g_strdup_printf("%" G_GINT64_FORMAT, h));
     g_hash_table_insert(osr->properties,
-			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LAYER_DOWNSAMPLE, i),
-			g_strdup(downsample));
+			g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_DOWNSAMPLE, i),
+			_openslide_format_double(osr->downsamples[i]));
+
+    // tile geometry
+    w = h = -1;
+    (osr->ops->get_tile_geometry)(osr, i, &w, &h);
+    bool have_geometry = (w > 0 && h > 0);
+    if (i == 0) {
+      should_have_geometry = have_geometry;
+    }
+    if (!openslide_get_error(osr) && have_geometry != should_have_geometry) {
+      g_warning("Inconsistent tile geometry hints between levels");
+    }
+    if (have_geometry) {
+      g_hash_table_insert(osr->properties,
+                          g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_TILE_WIDTH, i),
+                          g_strdup_printf("%" G_GINT64_FORMAT, w));
+      g_hash_table_insert(osr->properties,
+                          g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_LEVEL_TILE_HEIGHT, i),
+                          g_strdup_printf("%" G_GINT64_FORMAT, h));
+    }
   }
 
   // fill in names
@@ -364,12 +394,12 @@ void openslide_close(openslide_t *osr) {
 }
 
 
-void openslide_get_layer0_dimensions(openslide_t *osr,
-				     int64_t *w, int64_t *h) {
-  openslide_get_layer_dimensions(osr, 0, w, h);
+void openslide_get_level0_dimensions(openslide_t *osr,
+                                     int64_t *w, int64_t *h) {
+  openslide_get_level_dimensions(osr, 0, w, h);
 }
 
-void openslide_get_layer_dimensions(openslide_t *osr, int32_t layer,
+void openslide_get_level_dimensions(openslide_t *osr, int32_t level,
 				    int64_t *w, int64_t *h) {
   *w = -1;
   *h = -1;
@@ -378,28 +408,43 @@ void openslide_get_layer_dimensions(openslide_t *osr, int32_t layer,
     return;
   }
 
-  if (!layer_in_range(osr, layer)) {
+  if (!level_in_range(osr, level)) {
     return;
   }
 
-  (osr->ops->get_dimensions)(osr, layer, w, h);
+  (osr->ops->get_dimensions)(osr, level, w, h);
 }
 
+void openslide_get_layer0_dimensions(openslide_t *osr,
+                                     int64_t *w, int64_t *h) {
+  openslide_get_level0_dimensions(osr, w, h);
+}
+
+void openslide_get_layer_dimensions(openslide_t *osr, int32_t level,
+                                    int64_t *w, int64_t *h) {
+  openslide_get_level_dimensions(osr, level, w, h);
+}
+
+
 const char *openslide_get_comment(openslide_t *osr) {
   return openslide_get_property_value(osr, OPENSLIDE_PROPERTY_NAME_COMMENT);
 }
 
 
-int32_t openslide_get_layer_count(openslide_t *osr) {
+int32_t openslide_get_level_count(openslide_t *osr) {
   if (openslide_get_error(osr)) {
     return -1;
   }
 
-  return osr->layer_count;
+  return osr->level_count;
+}
+
+int32_t openslide_get_layer_count(openslide_t *osr) {
+  return openslide_get_level_count(osr);
 }
 
 
-int32_t openslide_get_best_layer_for_downsample(openslide_t *osr,
+int32_t openslide_get_best_level_for_downsample(openslide_t *osr,
 						double downsample) {
   if (openslide_get_error(osr)) {
     return -1;
@@ -411,30 +456,39 @@ int32_t openslide_get_best_layer_for_downsample(openslide_t *osr,
   }
 
   // find where we are in the middle
-  for (int32_t i = 1; i < osr->layer_count; i++) {
+  for (int32_t i = 1; i < osr->level_count; i++) {
     if (downsample < osr->downsamples[i]) {
       return i - 1;
     }
   }
 
   // too big, return last
-  return osr->layer_count - 1;
+  return osr->level_count - 1;
+}
+
+int32_t openslide_get_best_layer_for_downsample(openslide_t *osr,
+						double downsample) {
+  return openslide_get_best_level_for_downsample(osr, downsample);
 }
 
 
-double openslide_get_layer_downsample(openslide_t *osr, int32_t layer) {
-  if (openslide_get_error(osr) || !layer_in_range(osr, layer)) {
+double openslide_get_level_downsample(openslide_t *osr, int32_t level) {
+  if (openslide_get_error(osr) || !level_in_range(osr, level)) {
     return -1.0;
   }
 
-  return osr->downsamples[layer];
+  return osr->downsamples[level];
+}
+
+double openslide_get_layer_downsample(openslide_t *osr, int32_t level) {
+  return openslide_get_level_downsample(osr, level);
 }
 
 
 int openslide_give_prefetch_hint(openslide_t *osr G_GNUC_UNUSED,
 				 int64_t x G_GNUC_UNUSED,
 				 int64_t y G_GNUC_UNUSED,
-				 int32_t layer G_GNUC_UNUSED,
+				 int32_t level G_GNUC_UNUSED,
 				 int64_t w G_GNUC_UNUSED,
 				 int64_t h G_GNUC_UNUSED) {
   g_warning("openslide_give_prefetch_hint has never been implemented and should not be called");
@@ -449,7 +503,7 @@ void openslide_cancel_prefetch_hint(openslide_t *osr G_GNUC_UNUSED,
 static void read_region(openslide_t *osr,
 			cairo_t *cr,
 			int64_t x, int64_t y,
-			int32_t layer,
+			int32_t level,
 			int64_t w, int64_t h) {
   // save the old pattern, it's the only thing push/pop won't restore
   cairo_pattern_t *old_source = cairo_get_source(cr);
@@ -466,9 +520,9 @@ static void read_region(openslide_t *osr,
   // saturate those seams away!
   cairo_set_operator(cr, CAIRO_OPERATOR_SATURATE);
 
-  if (layer_in_range(osr, layer)) {
+  if (level_in_range(osr, level)) {
     // offset if given negative coordinates
-    double ds = openslide_get_layer_downsample(osr, layer);
+    double ds = openslide_get_level_downsample(osr, level);
     int64_t tx = 0;
     int64_t ty = 0;
     if (x < 0) {
@@ -485,7 +539,7 @@ static void read_region(openslide_t *osr,
 
     // paint
     if (w > 0 && h > 0) {
-      (osr->ops->paint_region)(osr, cr, x, y, layer, w, h);
+      (osr->ops->paint_region)(osr, cr, x, y, level, w, h);
     }
   }
 
@@ -514,7 +568,7 @@ static bool ensure_nonnegative_dimensions(openslide_t *osr, int64_t w, int64_t h
 void openslide_read_region(openslide_t *osr,
 			   uint32_t *dest,
 			   int64_t x, int64_t y,
-			   int32_t layer,
+			   int32_t level,
 			   int64_t w, int64_t h) {
   if (!ensure_nonnegative_dimensions(osr, w, h)) {
     return;
@@ -538,16 +592,16 @@ void openslide_read_region(openslide_t *osr,
   // 3. We would like to constrain the intermediate surface to a reasonable
   //    amount of RAM.
   const int64_t d = 4096;
-  double ds = openslide_get_layer_downsample(osr, layer);
+  double ds = openslide_get_level_downsample(osr, level);
   for (int64_t row = 0; !openslide_get_error(osr) && row < (h + d - 1) / d;
           row++) {
     for (int64_t col = 0; !openslide_get_error(osr) && col < (w + d - 1) / d;
             col++) {
       // calculate surface coordinates and size
-      int64_t sx = x + col * d * ds;     // layer 0 plane
-      int64_t sy = y + row * d * ds;     // layer 0 plane
-      int64_t sw = MIN(w - col * d, d);  // layer plane
-      int64_t sh = MIN(h - row * d, d);  // layer plane
+      int64_t sx = x + col * d * ds;     // level 0 plane
+      int64_t sy = y + row * d * ds;     // level 0 plane
+      int64_t sw = MIN(w - col * d, d);  // level plane
+      int64_t sh = MIN(h - row * d, d);  // level plane
 
       // create the cairo surface for the dest
       cairo_surface_t *surface;
@@ -565,7 +619,7 @@ void openslide_read_region(openslide_t *osr,
       cairo_surface_destroy(surface);
 
       // paint
-      read_region(osr, cr, sx, sy, layer, sw, sh);
+      read_region(osr, cr, sx, sy, level, sw, sh);
 
       // done
       _openslide_check_cairo_status_possibly_set_error(osr, cr);
@@ -583,13 +637,17 @@ void openslide_read_region(openslide_t *osr,
 void openslide_cairo_read_region(openslide_t *osr,
 				 cairo_t *cr,
 				 int64_t x, int64_t y,
-				 int32_t layer,
+				 int32_t level,
 				 int64_t w, int64_t h) {
   if (!ensure_nonnegative_dimensions(osr, w, h)) {
     return;
   }
 
-  read_region(osr, cr, x, y, layer, w, h);
+  if (openslide_get_error(osr)) {
+    return;
+  }
+
+  read_region(osr, cr, x, y, level, w, h);
 
   _openslide_check_cairo_status_possibly_set_error(osr, cr);
 }
@@ -608,7 +666,7 @@ const char *openslide_get_property_value(openslide_t *osr, const char *name) {
     return NULL;
   }
 
-  return (const char *) g_hash_table_lookup(osr->properties, name);
+  return g_hash_table_lookup(osr->properties, name);
 }
 
 const char * const *openslide_get_associated_image_names(openslide_t *osr) {
@@ -628,9 +686,8 @@ void openslide_get_associated_image_dimensions(openslide_t *osr, const char *nam
     return;
   }
 
-  struct _openslide_associated_image *img =
-    (struct _openslide_associated_image *) g_hash_table_lookup(osr->associated_images,
-							      name);
+  struct _openslide_associated_image *img = g_hash_table_lookup(osr->associated_images,
+								name);
   if (img) {
     *w = img->w;
     *h = img->h;
@@ -644,9 +701,8 @@ void openslide_read_associated_image(openslide_t *osr,
     return;
   }
 
-  struct _openslide_associated_image *img =
-    (struct _openslide_associated_image *) g_hash_table_lookup(osr->associated_images,
-							       name);
+  struct _openslide_associated_image *img = g_hash_table_lookup(osr->associated_images,
+								name);
   if (img) {
     // this function is documented to do nothing on failure, so we need an
     // extra memcpy
@@ -662,11 +718,6 @@ void openslide_read_associated_image(openslide_t *osr,
   }
 }
 
-void _openslide_set_background_color_property(GHashTable *ht,
-					      uint8_t r, uint8_t g, uint8_t b) {
-  g_return_if_fail(g_hash_table_lookup(ht,
-				       OPENSLIDE_PROPERTY_NAME_BACKGROUND_COLOR) == NULL);
-
-  g_hash_table_insert(ht, g_strdup(OPENSLIDE_PROPERTY_NAME_BACKGROUND_COLOR),
-		      g_strdup_printf("%.02X%.02X%.02X", r, g, b));
+const char *openslide_get_version(void) {
+  return PACKAGE_VERSION;
 }
